
#include "..\..\common\ImageFilter.h"


using namespace FilterSpace;
using namespace System::Windows::Forms;

bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// BEGIN --> COMMON SUPPORT CODE (DO NOT MODIFY!)
	sx = INTERVAL_LIMITER(sx, 0, f->width - 1);
	sy = INTERVAL_LIMITER(sy, 0, f->height - 1);
	if (sw < 1) sw = f->width - sx;
	if (sh < 1) sh = f->height - sy;
	sw = INTERVAL_LIMITER(sw, 1, f->width - sx);
	sh = INTERVAL_LIMITER(sh, 1, f->height - sy);
	//MessageBox::Show(sx.ToString() + " " + sy.ToString() + " " + sw.ToString() + " " + sh.ToString());

	int   Channels  = f->channelcountin;
	int   StrideIn  = f->stridein;
	int   StrideOut = f->strideout;
	byte *DataIn    = f->datain;
	byte *DataOut   = f->dataout;
	byte *mask      = f->mask;

	int x, y;
	byte *din, *dout;
	int processed_pixels = 0;

	int min[3] = {255, 255, 255};
	int max[3] = {  0,   0,   0};

	if (f->filter_pass == 1) // calculate statistics
	{
		for (y = 0; y < sh; ++y)
		{
			din  = f->datain  + (y + sy) * StrideIn + sx * Channels;
			for (x = 0; x < sw; ++x)
			{
				for (int t = 0; t < Channels; ++t) 
				{
					if (max[t] < din[t]) max[t] = din[t];
					if (min[t] > din[t]) min[t] = din[t];
				}
				din  += Channels;
			}

			// FOR PERFORMANCE MEASUREMENT
			processed_pixels += sw;
			ImageFilterBase::UpdateFilterStatus(f, processed_pixels, updater);
			if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
		}

		if (updater != nullptr) updater->WaitOne();

		for (int t = 0; t < Channels; ++t) 
		{
			if (max[t] > f->max[t]) f->max[t] = max[t];
			if (min[t] < f->min[t]) f->min[t] = min[t];
		}

		if (updater != nullptr) updater->Release(1);
	}
	else // normalize image
	{
		byte *mmin  = f->min;
		byte *mmax  = f->max;
		int   mw[3] = { f->max[0] - f->min[0], f->max[1] - f->min[1], f->max[2] - f->min[2] };
		
		for (y = 0; y < sh; ++y)
		{
			din  = f->datain  + (y + sy) * StrideIn + sx * Channels;
			dout = f->dataout +  y * StrideOut;

			for (x = 0; x < sw; ++x)
			{
				for (int t = 0; t < Channels; ++t) dout[t] = ((din[t] - mmin[t]) * 255) / mw[t];
				din  += Channels;
				dout += Channels;
			}

			// FOR PERFORMANCE MEASUREMENT
			processed_pixels += sw;
			ImageFilterBase::UpdateFilterStatus(f, processed_pixels, updater);
			if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
		}
		
	}

	return true;
}


//
//
IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int run_local = 1;

	if (!fs) return IntPtr(0);
	xml_node^ xn = gcnew xml_node( fs );

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	filtersettings* fset = XmlToMyStruct(xn, true);
	byte* indata  = fset->datain;
	byte* outdata = fset->dataout;
	int   width   = fset->width;
	int   height  = fset->height;
	int   stride  = fset->stridein;
	
	delete fset;

	xn->XmlRemoveChildByName("OutputImage0");
	xn->XmlGetChildByName(xt, "InputImage0");

	Byte* f = xt->XML_Tree_Marshal();
	xt = gcnew xml_node(f);
	delete [] f;
	xt->name = "OutputImage0";

	// SET "Run_Local" settings
	if (!run_local)
	{
		if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
		{
			if (stride * height == 0) return IntPtr(0);
			xt->XmlSetChildValueByName("Data", ((int)(void*)(new Byte[stride * height])));
		}
		else
		{
			xt->XmlSetChildValueByName("Data", (int)(void*)outdata);
		}
	}
	xn->children.Add( xt );
	return (IntPtr)(void*)xn->XML_Tree_Marshal();
}



filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	int       iTmp;
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;
	int packed, interleaved;
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) 
	{
		MessageBox::Show("XmlToMyStruct - 0 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, iTmp, iTmp, filset->strideout, iTmp, iTmp, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) ) ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) ) ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;


	return filset;
}



bool ImageFilter::CanLocalize(){ return true; }

IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            0,         0,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,           0,         0,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}




// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	if ((f->pixelsizein >> 3) << 3 != f->pixelsizein) return false;

	bool b;
	f->filter_pass = 1;
	f->max         = new byte[3];
	f->min         = new byte[3];
	memset(f->min, 255, 3);
	memset(f->max,   0, 3);

	if (b = (::FilterCore(f, sx, sy, sw, sh, nullptr)))
	{
		f->filter_pass = 2;
		b = (::FilterCore(f, sx, sy, sw, sh, nullptr));
	}
	delete [] f->min;
	delete [] f->max;
	return b;
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;

	if (filset)
	{
		if ((filset->pixelsizein >> 3) << 3 == filset->pixelsizein)
		{
			filset->max         = new byte[3];
			filset->min         = new byte[3];
			memset(filset->min, 255, 3);
			memset(filset->max,   0, 3);

			*(filset->processing_ptr)  = 0;
			*(filset->processing_size) = filset->width * filset->height;
			filset->nCoreExecutionSize = 1;
			filter_core_func f         = (::FilterCore);
			b                          = true;
			int cpucores               = System::Environment::ProcessorCount;

			if (filset->nCoreEnabled && cpucores == 1) filset->nCoreEnabled = 0;

			for (int i = 1; i <= 2; i++)
			{
				filset->filter_pass = i;
				if (b) b = (filset->nCoreEnabled) ? (ImageFilterBase::ExecuteFilterBlock(filset, f)) : (f(filset, 0, 0, 0, 0, nullptr));
			}
			
			delete [] filset->max;
			delete [] filset->min;
		}
		delete filset;
		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}


